import * as helpers from "./modules/helpers.js";
import { Point, Vector } from "./modules/Vector.js";

const config = {
  controls: {
    mouse: true,
    button: true,
    voice: false,
    motion: true,
  },
};

const navigation = {
  home: {
    label: "Home",
    path: "/",
    command: "home",
    submenu: {
      menu: {
        label: "Menu",
        path: "/menu",
        command: "menu",
        submenu: {
          call: {
            label: "Call",
            path: "/menu/call",
            command: "call",
            submenu: {
              emergency: {
                label: "Emergency",
                command: "emergency",
                path: "/menu/call/emergency",
                message: "Calling Emergency",
              },
              contacta: {
                label: "Flute Inc.",
                command: "contact a",
                path: "/menu/call/contacta",
                message: "Calling Flute Inc.",
              },
              contactb: {
                label: "Brays AG",
                command: "contact b",
                path: "/menu/call/contactb",
                message: "Calling Brays AG",
              },
            },
          },
          check: {
            label: "Check",
            path: "/menu/check",
            command: "check",
            submenu: {
              daily: {
                label: "Daily",
                command: "daily",
                path: "/menu/check/daily",
                message: "Displaying Daily Checklist",
              },
              weekly: {
                label: "Weekly",
                command: "weekly",
                path: "/menu/check/weekly",
                message: "Displaying Weekly Checklist",
              },
              monthly: {
                label: "Monthly",
                command: "monthly",
                path: "/menu/check/monthly",
                message: "Displaying Monthly Checklist",
              },
            },
          },
          apps: {
            label: "Apps",
            path: "/menu/apps",
            command: "apps",
            submenu: {
              record: {
                label: "Record",
                command: "record",
                path: "/menu/apps/record",
                message: "Opening Recording App",
              },
              breakout: {
                label: "Breakout",
                command: "breakout",
                path: "/menu/apps/breakout",
                message: "Opening Breakout App",
              },
              statistics: {
                label: "Statistics",
                command: "statistics",
                path: "/menu/apps/statistics",
                message: "Opening Statistics App",
              },
            },
          },
        },
      },
    },
  },
};

// Create WebSocket connection.
const socket = new WebSocket(`wss://192.168.227.52:443`);

// init application
function init() {
  // build navigation
  const appEl = document.getElementById("app");
  const navigationEl = document.createElement("nav");
  const listEl = document.createElement("ul");
  navigationEl.id = "navigation";
  navigationEl.appendChild(listEl);
  appEl.appendChild(navigationEl);

  // navigate to current path or root
  const pathname = window.location.pathname;
  if (pathname) navigateToPath(pathname);
  else navigateToPath("/");

  // Add a popstate event listener to the window object
  window.addEventListener("popstate", function (event) {
    var state = event.state;

    // Load the appropriate page based on the state object
    // without pushing it to the history
    if (state) navigateToPath(state.pageUrl, false);
    else navigateToPath("/", false);
  });

  // adding controls
  if (config.controls.mouse) addMouseControls();
  if (config.controls.button) addButtonControls();
  if (config.controls.voice) addVoiceControls();
  if (config.controls.motion && "RelativeOrientationSensor" in window)
    addMotionControls();

  // listen for WebSocket messages
  socket.addEventListener("message", (event) => {
    let message = JSON.parse(event.data);
    console.log(message);

    // perform action without broadcasting message again
    if (message.type === "select") navigateToPath(message.value, true, false);
    else if (message.type === "focus")
      focusNavigationItem(message.value, false);
  });
}

// update navigation
function buildPageContents(object) {
  // handle message content
  if (object.hasOwnProperty("message")) {
    // create content element
    const appEl = document.getElementById("app");
    const contentEl = document.createElement("div");
    contentEl.id = "content";
    contentEl.innerHTML = object.message;
    appEl.prepend(contentEl);
  } else {
    // remove content element
    const contentEl = document.getElementById("content");
    if (contentEl) contentEl.remove();
  }

  updateNavigation();

  // update navigation
  function updateNavigation() {
    const listEl = document.createElement("ul");
    const listItems = []; // for setting animation

    // add navigation items
    for (let key in object.submenu) {
      let itemEl = buildNavigationItemEl(object.submenu[key]);
      itemEl.classList.add("forward-btn");
      listEl.appendChild(itemEl);
      listItems.push(itemEl);
    }

    // add "special back" button
    if (object.path !== "/") {
      // determine depth of object
      const segments = object.path.split("/");
      const depth = segments.length - 1;
      let backItemEl;
      if (depth === 1)
        backItemEl = buildNavigationItemEl({
          label: "Close",
          path: "/",
          command: "close",
        });
      else if (depth === 2)
        backItemEl = buildNavigationItemEl({
          label: "Back",
          path: "/menu",
          command: "back",
        });
      else if (depth === 3) {
        backItemEl = buildNavigationItemEl({
          label: "Exit",
          path: "/",
          command: "exit",
        });
        backItemEl.classList.add("exit-btn");
      }
      backItemEl.classList.add("back-btn");
      listEl.appendChild(backItemEl);
      listItems.push(backItemEl);
    }

    // add animation delay
    let index = 0;
    listItems.forEach((item) => {
      item.style.animationDelay = index * 0.07 + "s";
      index++;
    });

    const navigationEl = document.getElementById("navigation");
    navigationEl.replaceChildren(listEl);

    // dispatch custom event to listen for menu updates
    const updated = new CustomEvent("updated");
    navigationEl.dispatchEvent(updated);

    // auto select first item
    // focusNavigationItem(1);

    // build navigation item
    function buildNavigationItemEl(item) {
      const itemEl = document.createElement("li");
      itemEl.className = "link";
      const linkEl = document.createElement("a");
      linkEl.href = item.path;
      linkEl.innerHTML = item.label;
      linkEl.dataset.command = item.command;
      itemEl.appendChild(linkEl);

      if (item.label.toLowerCase() !== item.command.toLowerCase()) {
        const commandEl = document.createElement("div");
        commandEl.className = "command";
        commandEl.innerHTML = item.command;
        itemEl.appendChild(commandEl);
      }

      return itemEl;
    }
  }
}

// general navigation
// ===================

// navigate to path
function navigateToPath(path, pushState = true, broadcast = true) {
  // find object with path
  let navObj = helpers.findNestedObject(navigation, "path", path);
  if (!navObj) navObj = navigation.home; // fallback to home
  buildPageContents(navObj);

  // State Object
  var stateObj = {
    pageTitle: navObj.label,
    pageUrl: navObj.path,
  };

  // Push state to browser history
  if (pushState)
    history.pushState(stateObj, stateObj.pageTitle, stateObj.pageUrl);

  // broadcast selection event to other clients
  if (broadcast && socket.readyState === WebSocket.OPEN)
    socket.send(JSON.stringify({ type: "select", value: path }));
}

// navigate to currently selected item
function navigateToSelected() {
  const currentItem = document.querySelector(
    "#navigation > ul > li[data-selected] a"
  );
  if (currentItem) {
    const path = currentItem.getAttribute("href");
    navigateToPath(path);
  } else printLogMsg("no item selected to open");
}

// focus specific navigation item
function focusNavigationItem(index, broadcast = true) {
  // unselect all items
  const navigationList = document.querySelectorAll("#navigation li");
  for (let item of navigationList) delete item.dataset.selected;

  // select requested element
  const navigationItem = document.querySelector(
    `#navigation > ul > li:nth-child(${index})`
  );
  navigationItem.dataset.selected = "";

  // broadcast focus event to other clients
  if (broadcast && socket.readyState === WebSocket.OPEN)
    socket.send(JSON.stringify({ type: "focus", value: index }));
}

// focus next navigation item
function focusNextNavigationItem() {
  // get currently selected element
  let currentItem = document.querySelector(
    "#navigation > ul > li[data-selected]"
  );

  if (!currentItem) focusNavigationItem(1);
  else if (currentItem.nextSibling) {
    delete currentItem.dataset.selected;
    currentItem.nextSibling.dataset.selected = "";
  } else focusNavigationItem(1);
}

// focus previous navigation item
function focusPreviousNavigationItem() {
  // get currently selected element
  let currentItem = document.querySelector(
    "#navigation > ul > li[data-selected]"
  );

  if (!currentItem) focusNavigationItem(1);
  else if (currentItem.previousSibling) {
    delete currentItem.dataset.selected;
    currentItem.previousSibling.dataset.selected = "";
  } else {
    let childCount = document.querySelector("#navigation > ul").children.length;
    focusNavigationItem(childCount);
  }
}

// focus closest item to point
function focusClosestItem(
  focusPoint = new Point(window.innerWidth / 2, window.innerHeight / 2)
) {
  let lastDistance = 10000;
  let index = 1;
  let closestIndex = 0;
  const itemList = document.querySelectorAll("#navigation ul li");
  for (let item of itemList) {
    let rect = item.getBoundingClientRect();
    let itemCenterPoint = new Point(
      rect.left + rect.width / 2,
      rect.top + rect.height / 2
    );
    let vector = new Vector(focusPoint, itemCenterPoint);
    if (vector.magnitude <= lastDistance) {
      closestIndex = index;
      lastDistance = vector.magnitude;
    }
    index++;
  }
  focusNavigationItem(closestIndex);
}

// moving up navigation tree
function moveUpNavigationLevel() {
  const path = window.location.pathname;
  let parentPath = path.slice(0, path.lastIndexOf("/"));
  navigateToPath(parentPath);
}

// moving down navigation tree
function moveDownNavigationLevel() {
  const currentItem = document.querySelector(
    "#navigation > ul > li[data-selected]"
  );
  if (currentItem) {
    const isBackButton = currentItem.classList.contains("back-btn");
    if (!isBackButton) navigateToSelected();
  }
}

// add mouse controls
// ==================
function addMouseControls() {
  const appEl = document.getElementById("app");

  // focus on mouseover
  appEl.addEventListener("mouseover", (event) => {
    const listItem = event.target.closest("li.link");
    if (listItem) {
      // get childnode index
      let index = Array.from(listItem.parentNode.children).indexOf(listItem);
      focusNavigationItem(index + 1);
    }
  });

  // unfocus all items
  appEl.addEventListener("mouseout", (event) => {
    const listItem = event.target.closest("li.link");
    if (listItem) {
      // unselect all items
      const navigationList = document.querySelectorAll("#navigation li");
      for (let item of navigationList) delete item.dataset.selected;
    }
  });

  // select on mouseclick
  appEl.addEventListener("click", (event) => {
    event.preventDefault();
    const listItem = event.target.closest("li.link");
    if (listItem) {
      const link = listItem.querySelector("a");
      const path = link.getAttribute("href");
      navigateToPath(path);
    }
  });
}

// add button controls
// ===================
function addButtonControls() {
  // add two button controls for menu
  document.addEventListener("keydown", onKeyDown);

  // on keydown event
  function onKeyDown(event) {
    // prevent defaults
    if (
      event.code === "Tab" ||
      event.code === "Enter" ||
      event.code === "Escape" ||
      event.code === "ArrowUp" ||
      event.code === "ArrowRight" ||
      event.code === "ArrowDown" ||
      event.code === "ArrowLeft"
    )
      event.preventDefault();

    // browsing navigation items
    if (event.code === "Tab" || event.code === "ArrowDown")
      focusNextNavigationItem();
    else if (event.code === "ArrowUp") focusPreviousNavigationItem();
    else if (event.code === "Enter") navigateToSelected();
    else if (event.code === "ArrowRight") moveDownNavigationLevel();
    else if (event.code === "ArrowLeft" || event.code === "Escape")
      moveUpNavigationLevel();
    else if (event.code === "VolumeUp") printLogMsg("yes");
  }

  // const toggleButtonControlsEl = document.getElementById(
  //   "toggle-button-controls"
  // );

  // toggleButtonControlsEl.addEventListener("click", (event) => {
  //   const audioElement = document.createElement("audio");
  //   const audioCtx = new AudioContext();
  //   const source = audioCtx.createMediaElementSource(audioElement);
  //   const gainNode = audioCtx.createGain();
  //   source.connect(gainNode).connect(audioCtx.destination);

  //   // Set the initial volume level
  //   gainNode.gain.value = 0.2;
  //   console.log(gainNode.gain.value);

  // Adjust the volume when the device volume buttons are pressed
  // document.addEventListener("keydown", (event) => {
  //   if (event.key === "ArrowUp") {
  //     printLogMsg("volume up");
  //     gainNode.gain.value += 0.1;
  //   } else if (event.key === "ArrowDown") {
  //     printLogMsg("volume down");
  //     gainNode.gain.value -= 0.1;
  //   }
  // });
  // });
}

// add voice controls
// ===================
function addVoiceControls() {
  // voice controls only work on google chrome!
  var SpeechRecognition = SpeechRecognition || webkitSpeechRecognition;
  var SpeechGrammarList = SpeechGrammarList || window.webkitSpeechGrammarList;
  var SpeechRecognitionEvent =
    SpeechRecognitionEvent || webkitSpeechRecognitionEvent;

  let availableCommands = [];
  let commands = [];
  updateVoiceCommandList();

  // listen for menu updates and refresh commands list
  const navigationEl = document.getElementById("navigation");
  navigationEl.addEventListener("updated", (event) => {
    updateVoiceCommandList();
  });

  let recognition = new SpeechRecognition();
  if (SpeechGrammarList) {
    var speechRecognitionList = new SpeechGrammarList();
    var grammar =
      "#JSGF V1.0; grammar commands; public <command> = " +
      commands.join(" | ") +
      " ;";
    speechRecognitionList.addFromString(grammar, 1);
    // recognition.grammars = speechRecognitionList;
  }
  recognition.continuous = false;
  recognition.lang = "en-US";
  recognition.interimResults = false;
  recognition.maxAlternatives = 1;

  recognition.start();
  printLogMsg("Voice > listening");

  recognition.onresult = function (event) {
    let command = event.results[0][0].transcript.toLowerCase();
    let confidence = event.results[0][0].confidence;
    command = command.toLowerCase();

    printLogMsg(`Voice > Received "${command}" (${confidence.toFixed(2)})`);

    if (commands.includes(command)) {
      let index = commands.indexOf(command); // get index of command in array
      let path = availableCommands[index].path; // get corresponding path in availableCommands array
      navigateToPath(path);
    }
  };

  recognition.onspeechend = function () {
    recognition.stop();

    setTimeout(() => {
      recognition.start();
      printLogMsg("Voice > listening");
    }, "1000");
  };

  recognition.onnomatch = function (event) {
    printLogMsg("I didn't recognise that command.");
  };

  // updating voice command list
  function updateVoiceCommandList() {
    // search and save currently available voice commands
    availableCommands = [];
    const listItems = document.querySelectorAll("a[data-command]");
    for (let item of listItems) {
      let command = {
        command: item.dataset.command,
        path: item.getAttribute("href"),
      };
      availableCommands.push(command);
    }

    // add custom commands
    availableCommands.push({ command: "exit", path: "/" });
    availableCommands.push({
      command: "emergency call",
      path: "/menu/call/emergency",
    });

    // form commands array for grammarlist
    commands = availableCommands.map((x) => x.command);
    printLogMsg(`Voice > Commands: [${commands}]`);
  }
}

// add motion controls
// ===================
function addMotionControls() {
  let pointOfReference;
  let vectorToReference;
  let screenCenter;

  // create canvas element for motion visualization
  const canvasEl = document.createElement("canvas");
  canvasEl.id = "canvas";
  canvasEl.width = window.innerWidth;
  canvasEl.height = window.innerHeight;
  document.getElementById("app").appendChild(canvasEl);

  //available headset sensors: accelerometer, gyroscope, relativeOrientation, linearAcceleration, GravitySensor
  //unavailable headset sensors: magnetometer (this results in horizontal drifting)
  const options = { frequency: 20, referenceFrame: "device" };
  const sensor = new RelativeOrientationSensor(options);
  sensor.addEventListener("reading", onChangeOrientation);

  // get sensor permission via Permission API
  Promise.all([
    navigator.permissions.query({ name: "accelerometer" }),
    navigator.permissions.query({ name: "gyroscope" }),
  ]).then((results) => {
    if (results.every((result) => result.state === "granted")) sensor.start();
    else printLogMsg("No permissions to use RelativeOrientationSensor.");
  });

  // handle motion sensor event
  function onChangeOrientation(event) {
    let quaternion = event.target.quaternion;
    if (pointOfReference === undefined) setPointOfReference(quaternion);

    let mapped = {
      x: quaternion[0],
      y: quaternion[1],
      z: quaternion[2],
      w: quaternion[3],
    };

    let targetPoint = new Point(mapped.x, mapped.y);
    let motionVector = new Vector(pointOfReference, targetPoint);
    motionVector = motionVector.subtract(vectorToReference);
    motionVector = motionVector.scale(200);
    motionVector = motionVector.add(
      new Vector(
        new Point(),
        new Point(canvasEl.width / 2, canvasEl.height / 2)
      )
    );

    // focus closest item
    // ADD THRESHOLD BEFORE TRIGGERING!
    focusClosestItem(motionVector);

    drawMotion(motionVector);
    // printLogMsg(`${motionVector.magnitude.toFixed(2)}`);
  }

  // set new point of reference
  function setPointOfReference(quaternion) {
    pointOfReference = new Point(quaternion[0], quaternion[1]);
    vectorToReference = new Vector(new Point(), pointOfReference);
  }

  function drawMotion(vector) {
    if (canvasEl.getContext) {
      const ctx = canvasEl.getContext("2d");
      const pointSize = 10;

      ctx.clearRect(0, 0, canvasEl.width, canvasEl.height);

      ctx.strokeStyle = "#FFFFFF";
      ctx.fillStyle = "#FFFFFF";
      ctx.lineWidth = 2;

      ctx.beginPath();
      ctx.moveTo(canvasEl.width / 2, canvasEl.height / 2);
      ctx.lineTo(vector.x, vector.y);
      ctx.stroke();
      ctx.fillRect(
        vector.x - pointSize / 2,
        vector.y - pointSize / 2,
        pointSize,
        pointSize
      );
    }
  }
}

// printing messages on screen for debugging
function printLogMsg(message) {
  const logEl = document.getElementById("log");
  if (logEl.scrollHeight >= 10000) logEl.innerHTML = "";
  logEl.innerHTML += `${message}<br>`;
  logEl.scrollTop = logEl.scrollHeight;
  console.log(message);
}

init();
